#!/usr/bin/env python
#
# This file is part of vospace-transfer-service
# Copyright (C) 2021 Istituto Nazionale di Astrofisica
# SPDX-License-Identifier: GPL-3.0-or-later
#

import json
import logging
import os

from config import Config
from db_connector import DbConnector
from mailer import Mailer
from redis_log_handler import RedisLogHandler
from task_executor import TaskExecutor


class DOIPreprocessor(TaskExecutor):

    def __init__(self):
        self.type = "retrieve_preprocessor"
        config = Config("/etc/vos_ts/vos_ts.conf")
        params = config.loadSection("mail")
        self.adminEmail = params["admin_email"]
        params = config.loadSection("logging")
        self.logger = logging.getLogger(__name__)
        logLevel = "logging." + params["log_level"]
        logFormat = params["log_format"]
        logFormatter = logging.Formatter(logFormat)
        self.logger.setLevel(eval(logLevel))
        redisLogHandler = RedisLogHandler()
        redisLogHandler.setFormatter(logFormatter)
        self.logger.addHandler(redisLogHandler)
        params = config.loadSection("file_catalog")
        self.dbConn = DbConnector(params["user"],
                                  params["password"],
                                  params["host"],
                                  params.getint("port"),
                                  params["db"],
                                  1,
                                  1,
                                  self.logger)
        self.jobObj = None
        self.nodeList = []
        super(DOIPreprocessor, self).__init__()

    def execute(self):
        self.logger.info("Generating VOSpace node list")
        target = self.jobObj.jobInfo["transfer"]["target"].split("!vospace")[1]
        params = self.jobObj.jobInfo["transfer"]["view"]["param"]
        if not params:
            self.nodeList.append(target)
        else:
            for el in params:
                self.nodeList.append(target + '/' + el["value"])
        self.jobObj.nodeList = self.nodeList.copy()
        self.dbConn.insertJob(self.jobObj)

    def update(self, status):
        m = Mailer(self.logger)
        m.addRecipient(self.adminEmail)
        userEmail = self.dbConn.getUserEmail(self.jobObj.ownerId)
        if userEmail != self.adminEmail:
            m.addRecipient(userEmail)

        msg = f"""
        Dear user,
        your job has been {status}.

        Job ID: {self.jobObj.jobId}
        Job type: {self.jobObj.type}
        Owner ID: {self.jobObj.ownerId}

        """

        # Send e-mail notification
        m.setMessage(f"VOSpace DOI notification: Job {status}", msg)
        m.send()

    def cleanup(self):
        self.nodeList.clear()
        self.setDestinationQueueName("doi_ready")

    def run(self):
        self.logger.info("Starting DOI preprocessor...")
        self.setSourceQueueName("doi_pending")
        self.setDestinationQueueName("doi_ready")
        while True:
            self.wait()
            try:
                srcQueueLen = self.srcQueue.len()
                destQueueLen = self.destQueue.len()
            except Exception:
                self.logger.exception("Cache error: unable to retrieve queue length.")
            else:
                if destQueueLen < self.maxReadyJobs and srcQueueLen > 0:
                    self.jobObj = self.srcQueue.getJob()
                    jobId = self.jobObj.jobId
                    self.execute()
                    try:
                        jobPhase = self.dbConn.getJobPhase(jobId)
                    except Exception:
                        self.logger.exception(f"Database error: unable to retrieve job phase for job {jobId}.")
                    else:
                        if jobPhase == "ABORTED":
                            self.jobObj.setPhase("ABORTED")
                            self.setDestinationQueueName("read_terminated")
                            self.update("ABORTED")
                        else:
                            self.update("QUEUED")
                        try:
                            self.destQueue.insertJob(self.jobObj)
                            self.srcQueue.extractJob()
                        except Exception:
                            self.logger.exception(f"Failed to move job {self.jobObj.jobId} from '{self.srcQueue.name()}' to '{self.destQueue.name()}'")
                        else:
                            self.logger.info(f"Job {self.jobObj.jobId} MOVED from '{self.srcQueue.name()}' to '{self.destQueue.name()}'")
            finally:
                self.cleanup()
