#!/usr/bin/env python
#
# This file is part of vospace-transfer-service
# Copyright (C) 2021 Istituto Nazionale di Astrofisica
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
import shutil
import subprocess
import sys

from exceptions import TarFileCreationException


class FileGrouper(object):

    def __init__(self, minNumOfFiles, maxFolderSize):
        """
        Constructor requiring the minimum number of files
        and the maximum folder size in bytes (constraints).
        """
        self.minNumOfFiles = minNumOfFiles
        self.maxFolderSize = maxFolderSize

    def setMinNumOfFiles(self, minNumOfFiles):
        """Sets the 'minimum number of files' constraint."""
        self.minNumOfFiles = minNumOfFiles
      
    def setMaxFolderSize(self, maxFolderSize):
        """Sets the 'maximum folder size' constraint."""
        self.maxFolderSize = maxFolderSize

    def getLeaves(self):
        """
        Returns a list of all the 'leaves' (directories not
        containing subdirectories) within the path.
        """
        leaves = []
        for root, dirs, files in os.walk(self.folderPath):
            if not dirs:
                leaves.append(root)
        return leaves

    def getFolderSize(self, folder):
        """Returns the folder size in bytes."""
        files = []
        for f in os.listdir(folder):
            files.append(os.path.abspath(folder) + '/' + f)  
        folderSize = 0
        for f in files:
            folderSize += os.path.getsize(f)
        return folderSize

    def getNumOfFiles(self, folder):
        """Returns the number of files contained into a folder."""
        return len(os.listdir(folder))

    def recursive(self, folderPath):
        """Starts the .tar creation procedure."""
        self.folderPath = folderPath
        leaves = self.getLeaves()
        for folder in leaves:
            if(self.getNumOfFiles(folder) > self.minNumOfFiles and self.getFolderSize(folder) < self.maxFolderSize):
                cwd = os.getcwd()
                parent = os.path.dirname(folder)
                os.chdir(parent)
                try:
                    sp = subprocess.run(["tar", "-cf", os.path.basename(folder) + ".tar", os.path.basename(folder)], capture_output = True)
                except Exception:
                    raise
                else:
                    if(sp.returncode or sp.stderr):
                        raise(TarFileCreationException(folder))
                    else:
                        try:
                            shutil.rmtree(folder)
                        except Exception:
                            raise
                    os.chdir(cwd)


# Test (creates a .tar of all the leaves containing at least 2 files and having size lower than 100G,
# than removes those directories)
#fg = FileGrouper(1, 100 * (2 ** 30))
#fg.recursive("/home/cristiano/Lavoro/md5_test/store/a")
